/*

PhyML:  a program that  computes maximum likelihood phylogenies from
DNA or AA homologous sequences.

Copyright (C) Stephane Guindon. Oct 2003 onward.

All parts of the source except where indicated are distributed under
the GNU public licence. See http://www.opensource.org for details.

*/

#include <config.h>

#ifndef OPTIONS_H
#define OPTIONS_H

#include "utilities.h"
#include "cl.h"
#include "models.h"
#include "free.h"
#include "interface.h"

void Usage();
#endif
